import hashlib
import re

from pysatl import Utils

from sha256bit import Sha256bit


def block_generator(seed, msg_bitlen):
    l2block = hashlib.sha256(seed).digest()
    l2block += hashlib.sha256(l2block).digest()
    block_bitlen = len(l2block) * 8
    assert block_bitlen == 64 * 8
    bitlen = 0
    while bitlen + block_bitlen < msg_bitlen:
        yield l2block
        bitlen += block_bitlen
        l2block = hashlib.sha256(l2block).digest()
        l2block += hashlib.sha256(l2block).digest()
    # last block
    last_block_bitlen = msg_bitlen % 512
    if 0 != last_block_bitlen:
        last_block_full_bytelen = (last_block_bitlen + 7) // 8
        l2block = bytearray(l2block[0:last_block_full_bytelen])
        assert len(l2block) == last_block_full_bytelen
        mask = 0xFF & (0xFF << (8 - (msg_bitlen % 8)))
        if 0 != mask:
            l2block[-1] &= mask
    yield l2block


def msg_generator(seed, msg_bitlen):
    o = bytearray()
    for b in block_generator(seed, msg_bitlen):
        o += b
    return o


def check_against_hashlib(n_seeds=3, max_length=1024 * 4):
    print('check against hashlib')

    assert hashlib.sha256(b'abc').digest() == Sha256bit(b'abc').digest()

    def check_against_hashlib(seed, msg_bitlen):
        expected = hashlib.sha256()
        dut = Sha256bit()
        for block in block_generator(seed, msg_bitlen):
            # print(Utils.hexstr(block))
            expected.update(block)
            dut.update(block)
        assert expected.digest() == dut.digest()

    for seed_byte in range(0, n_seeds):
        for msg_bitlen in range(0, max_length, 8):
            seed = bytearray([seed_byte])
            check_against_hashlib(seed, msg_bitlen)


def check(msg, bitlen, sig):
    m = Sha256bit()
    if isinstance(msg, str):
        msg = msg.encode('ascii')
    descr = 'msg      = ' + Utils.hexstr(msg) + '\n'
    descr += 'bitlen   = %d\n' % bitlen
    descr += 'expected = ' + sig + '\n'
    try:
        m.update(msg, bitlen=bitlen)
        digest = m.hexdigest()
    except Exception as e:
        print(descr)
        raise e
    err_msg = '\n'
    err_msg += descr
    err_msg += 'digest   = ' + digest + '\n'
    assert digest == sig, err_msg


def check_hardcoded_test_vectors():
    print('check few minimal hardcoded test vectors')

    tests = [
        {
            'msg': '',
            'bitlen': 0,
            'digest': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
        },
        {
            'msg': 'a',
            'bitlen': 8,
            'digest': 'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb',
        },
        {
            'msg': Utils.ba('00'),
            'bitlen': 1,
            'digest': 'bd4f9e98beb68c6ead3243b1b4c7fed75fa4feaab1f84795cbd8a98676a2a375',
        },
        {
            'msg': Utils.ba('80'),
            'bitlen': 2,
            'digest': '18f331f626210ff9bad6995d8cff6e891adba50eb2fdbddcaa921221cdc333ae',
        },
    ]

    for test in tests:
        check(test['msg'], test['bitlen'], test['digest'])

    assert (
        Sha256bit(b'\x00', bitlen=1).hexdigest() == 'bd4f9e98beb68c6ead3243b1b4c7fed75fa4feaab1f84795cbd8a98676a2a375'
    )


def check_against_nist_cavp():
    print("check against 'short' and 'long' bit oriented test vectors from NIST CAVP")
    # (https://csrc.nist.gov/CSRC/media/Projects/Cryptographic-Algorithm-Validation-Program/documents/shs/shabittestvectors.zip)

    from pathlib import Path

    resource_path = Path(__file__).parent
    for tv_file in ['SHA256ShortMsg.rsp', 'SHA256LongMsg.rsp']:
        tv_path = resource_path.joinpath(tv_file)
        with open(tv_path) as f:
            for line in f:
                if line.startswith('Len'):
                    bitlen = int(re.search(r'Len = (.+)', line).group(1))
                if line.startswith('Msg'):
                    msg = Utils.ba(re.search(r'Msg = (.+)', line).group(1))
                    if bitlen == 0:
                        msg = bytes(0)
                if line.startswith('MD'):
                    md = re.search(r'MD = (.+)', line).group(1)
                    check(msg, bitlen, md)


def check_api():
    print('check API')
    msg = msg_generator(bytes(0), 300 * 8)
    expected = hashlib.sha256(msg).digest()
    # print(Utils.hexstr(msg))
    # print(Utils.hexstr(expected))
    assert expected == Sha256bit(msg).digest()
    for len1 in range(0, len(msg) * 8):
        dut1 = Sha256bit()
        dut1.update(msg[:len1])
        state = dut1.export_state()
        dut2 = Sha256bit.import_state(state)
        dut2.update(msg[len1:])
        assert expected == dut2.digest()
    for len1 in range(1, len(msg) * 8):
        dut = Sha256bit()
        remaining = len(msg)
        p = 0
        while remaining > 0:
            chunk = msg[p : p + len1]
            dut.update(chunk, bitlen=len(chunk) * 8)
            p += len1
            remaining -= len1
        assert expected == dut.digest()
        state = dut.export_state()
        dut2 = Sha256bit.import_state(state)
        assert expected == dut2.digest()
    dut = Sha256bit(b'\x00', bitlen=1)
    state = dut.export_state()
    dut2 = Sha256bit.import_state(state)
    assert dut2.hexdigest() == 'bd4f9e98beb68c6ead3243b1b4c7fed75fa4feaab1f84795cbd8a98676a2a375'


if __name__ == '__main__':
    check_api()
    check_hardcoded_test_vectors()
    check_against_nist_cavp()
    check_against_hashlib(n_seeds=3, max_length=1024 * 4)
    print('All test PASS')



import binascii
import copy
import struct


class Sha256bit:
    F32 = 0xFFFFFFFF

    _k = [
        0x428A2F98,
        0x71374491,
        0xB5C0FBCF,
        0xE9B5DBA5,
        0x3956C25B,
        0x59F111F1,
        0x923F82A4,
        0xAB1C5ED5,
        0xD807AA98,
        0x12835B01,
        0x243185BE,
        0x550C7DC3,
        0x72BE5D74,
        0x80DEB1FE,
        0x9BDC06A7,
        0xC19BF174,
        0xE49B69C1,
        0xEFBE4786,
        0x0FC19DC6,
        0x240CA1CC,
        0x2DE92C6F,
        0x4A7484AA,
        0x5CB0A9DC,
        0x76F988DA,
        0x983E5152,
        0xA831C66D,
        0xB00327C8,
        0xBF597FC7,
        0xC6E00BF3,
        0xD5A79147,
        0x06CA6351,
        0x14292967,
        0x27B70A85,
        0x2E1B2138,
        0x4D2C6DFC,
        0x53380D13,
        0x650A7354,
        0x766A0ABB,
        0x81C2C92E,
        0x92722C85,
        0xA2BFE8A1,
        0xA81A664B,
        0xC24B8B70,
        0xC76C51A3,
        0xD192E819,
        0xD6990624,
        0xF40E3585,
        0x106AA070,
        0x19A4C116,
        0x1E376C08,
        0x2748774C,
        0x34B0BCB5,
        0x391C0CB3,
        0x4ED8AA4A,
        0x5B9CCA4F,
        0x682E6FF3,
        0x748F82EE,
        0x78A5636F,
        0x84C87814,
        0x8CC70208,
        0x90BEFFFA,
        0xA4506CEB,
        0xBEF9A3F7,
        0xC67178F2,
    ]

    _h_init = [
        0x6A09E667,
        0xBB67AE85,
        0x3C6EF372,
        0xA54FF53A,
        0x510E527F,
        0x9B05688C,
        0x1F83D9AB,
        0x5BE0CD19,
    ]

    @staticmethod
    def _rotr(x, y):
        return ((x >> y) | (x << (32 - y))) & Sha256bit.F32

    @staticmethod
    def _maj(x, y, z):
        return (x & y) ^ (x & z) ^ (y & z)

    @staticmethod
    def _ch(x, y, z):
        return (x & y) ^ ((~x) & z)

    _output_size = 8
    blocksize = 1
    block_size = 64
    digest_size = 32

    def __init__(self, m=None, *, bitlen=None):
        """SHA-256 implementation supporting bit granularity for message input length.
        API is the same as hashlib.
        """
        self._counter = 0
        self._cache = bytearray()
        self._h = copy.deepcopy(Sha256bit._h_init)
        self._has_bitlen = False
        self._digest = None
        self.update(m, bitlen=bitlen)

    def export_state(self):
        if self._digest is None:
            h = self._h
            c = self._cache
        else:
            h = self._digest
            c = None
        return {'h': h, 'cnt': self._counter, 'cache': c}

    @staticmethod
    def import_state(state):
        o = Sha256bit()
        o._counter = state['cnt']
        if 0 != (o._counter % 8):
            o._has_bitlen = True
        if state['cache'] is None:
            o._digest = state['h']
            o._h = None
            o._cache = None
        else:
            o._h = state['h']
            o._cache = bytearray(state['cache'])
        return o

    def _compress(self, c):
        w = [0] * 64
        w[0:16] = struct.unpack('!16L', c)
        for i in range(16, 64):
            s0 = Sha256bit._rotr(w[i - 15], 7) ^ Sha256bit._rotr(w[i - 15], 18) ^ (w[i - 15] >> 3)
            s1 = Sha256bit._rotr(w[i - 2], 17) ^ Sha256bit._rotr(w[i - 2], 19) ^ (w[i - 2] >> 10)
            w[i] = (w[i - 16] + s0 + w[i - 7] + s1) & Sha256bit.F32

        a, b, c, d, e, f, g, h = self._h

        for i in range(64):
            s0 = Sha256bit._rotr(a, 2) ^ Sha256bit._rotr(a, 13) ^ Sha256bit._rotr(a, 22)
            t2 = s0 + Sha256bit._maj(a, b, c)
            s1 = Sha256bit._rotr(e, 6) ^ Sha256bit._rotr(e, 11) ^ Sha256bit._rotr(e, 25)
            t1 = h + s1 + Sha256bit._ch(e, f, g) + Sha256bit._k[i] + w[i]

            h = g
            g = f
            f = e
            e = (d + t1) & Sha256bit.F32
            d = c
            c = b
            b = a
            a = (t1 + t2) & Sha256bit.F32

        for i, (x, y) in enumerate(zip(self._h, [a, b, c, d, e, f, g, h])):
            self._h[i] = (x + y) & Sha256bit.F32

    def update(self, m, *, bitlen=None):
        """Update the hash object with the bytes in data. Repeated calls
        are equivalent to a single call with the concatenation of all
        the arguments.
        """
        if not m:
            return
        if self._has_bitlen:
            raise AssertionError('we support bitlen only for last call')
        bytes_bitlen = len(m) * 8
        if bitlen is not None:
            if 0 != (bitlen % 8):
                self._has_bitlen = True
                if bytes_bitlen - bitlen >= 8 or bitlen >= bytes_bitlen:
                    raise AssertionError(
                        'bitlen=%d, bytes_bitlen=%d'
                        % (
                            bitlen,
                            bytes_bitlen,
                        )
                    )
            else:
                if bitlen != bytes_bitlen:
                    raise AssertionError(
                        'bitlen=%d, bytes_bitlen=%d'
                        % (
                            bitlen,
                            bytes_bitlen,
                        )
                    )
            self._counter += bitlen
        else:
            self._counter += bytes_bitlen

        self._cache += m

        while len(self._cache) > 64:
            self._compress(self._cache[:64])
            self._cache = self._cache[64:]

        if len(self._cache) == 64 and 0 == (self._counter % 8):
            self._compress(self._cache[:64])
            self._cache = self._cache[64:]

    def _pad(self):
        last_block_bitlen = self._counter % 512
        last_block_full_bytes_cnt = last_block_bitlen // 8
        if last_block_bitlen < 448:
            padlen = 55 - last_block_full_bytes_cnt
        else:
            padlen = 119 - last_block_full_bytes_cnt

        shift = self._counter % 8
        if shift > 0 and (len(self._cache) > 0):
            mask = 0xFF << (8 - shift)
            self._cache[-1] = (self._cache[-1] & mask) | (0x80 >> shift)
        else:
            self._cache += b'\x80'
        self._cache += (b'\x00' * padlen) + self._counter.to_bytes(8, byteorder='big')

        if len(self._cache) not in [64, 128]:
            raise AssertionError('len(self._cache)=%d' % len(self._cache))

    def digest(self):
        """Return the digest of the bytes passed to the update() method
        so far as a bytes object.
        """
        if self._digest is not None:
            return self._digest
        self._pad()
        blocks = [self._cache[i : i + 64] for i in range(0, len(self._cache), 64)]
        for b in blocks:
            self._compress(b)
        data = [struct.pack('!L', i) for i in self._h[: self._output_size]]
        self._digest = b''.join(data)
        self._cache = None
        self._h = None
        return self._digest

    def hexdigest(self):
        """Like digest() except the digest is returned as a string
        of double length, containing only hexadecimal digits.
        """
        return binascii.hexlify(self.digest()).decode('ascii')
